import os
import glob
import csv

WHITE = (255,255,255)
BLACK = (0,0,0)
BLUE = (0,0,255)
REDGREENOREGON = (123,321,231)

DEFAULT_TRAVELLING_HOP = 0.05
DEFAULT_SCREEN_DIM = (1300,750)
DEFAULT_POINT_CACHE_SIZE = 10

# TODO:
def clear_and_make_directory(dirPath):
    """
    clear_and_make_directory

    :param dirPath: directory path
    :type dirPath: str or None
    :return: None
    :rtype: None
    """

    ##dirr = os.path.dirname(fullPath)
    if not os.path.isdir(dirPath):
        # make directory
        os.mkdir(dirPath)
    else:
        # TODO: clear directory contents
        files = glob.glob(dirPath + "/*")
        for f in files:
            try:
                os.remove(f)
            except:
                print("probably is folderonos")

def make_csv_file(filePath, columnLabels):
    with open(filePath, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quotechar='|', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(columnLabels)

def is_valid_color(c):
    assert type(c) is tuple, "invalid color type"
    for c_ in c: assert type(c_) in [float,int], "invalid type for colores de la vegasvalueorios"



from morebs2 import numerical_generator
import numpy as np
from collections import OrderedDict

import unittest

class TestNumericalGeneratorClass(unittest.TestCase):


    def test_CycleMap__random_map(self):
        vr = numerical_generator.CycleMap.random_cycle_map(5)
        ##print("VR")
        ##print(vr)

        f = numerical_generator.CycleMap.is_valid_map(vr)
        ##print("F: ", f)
        assert f, "random cycle map is not cycle"

        q = OrderedDict()
        x = [(3,1),(1,3),(2,4),(4,2)]
        for k,v in x:
            q[k] = v

        f2 = numerical_generator.CycleMap.is_valid_map(q)
        ##print("F2: ", f2)
        assert not f2, "non-cyclic map"
        return

    def test__generate_possible_binary_sequences(self):
        g = numerical_generator.generate_possible_binary_sequences(5, [])
        g = list(g)
        assert len(g) == 2 ** 5, "incorrect generation"

        # uncomment for viewing
        for g_ in g:
            print(g_)

    def test__random_npoint_from_point_in_bounds(self):

        bounds = np.array([[-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0]])
        r = 200.0

        # case point on >= 1 of the bound extremes
        p1 = np.array([-100,100,-99,80,95.0,-100])
        for i in range(100):
            p2 = numerical_generator.random_npoint_from_point_in_bounds(bounds,p1,r)
            assert type(p2) != type(None)
        return

    def test__random_npoint_from_point(self):
        r = 200.0

        # case point on >= 1 of the bound extremes
        p1 = np.array([-100,100,-99,80,95.0,-100])
        for i in range(100):
            p2 = numerical_generator.random_npoint_from_point(p1,r)
            assert type(p2) != type(None)
        return

######################## START: cycle map

def test_CycleMap__set_map():
    f = CycleMap.random_cycle_map(5)
    cm = CycleMap(5)
    cm.set_map(f)
    return

###########################


def test_FloatDeltaGenerator__next__():
    # case 1
    func = DELTA_ADD
    mox = round(1/7,5)
    fdg = FloatDeltaGenerator(0.0, DELTA_ADD, mox,1.0)

    c = 10
    while c > 0:
        q = next(fdg)
        print("r: ", q)
        if q == None: break
        c -= 1

    print("CELAS ", c)

    # case 2
    fdg = FloatDeltaGenerator_pattern1((1.0,0.0), 15.0)

    c = 20
    while c > 0:
        q = next(fdg)
        print("Q: ", q)
        c -= 1
        if q == None: break

    print("CELAS ", c)


if __name__ == '__main__':
    unittest.main()



'''
(a(X_n) + c) modulo m
'''
import random
from math import ceil
from .matrix_methods import *
from .variance_works import *
from collections import OrderedDict

class LCG:
    '''
    linear congruential generator
    '''
    
    def __init__(self, startingInteger, modulo):
        assert type(modulo) is int and modulo > 128, "invalid modulo"
        assert type(startingInteger) is int and startingInteger < modulo, "invalid starting integer"

        self.startingInteger = startingInteger
        self.vInt = startingInteger
        self.modulo = modulo
        self.fetch_arguments()

    # TODO: mod here?
    def fetch_arguments(self):
        self.multiplier = self.fetch_random_integer_in_range((0, self.modulo))
        self.increment = self.fetch_random_integer_in_range((0, self.modulo))

    def fetch_random_integer_in_range(self, ranje):
        assert type(ranje[0]) is int and type(ranje[1]) == type(ranje[0]), "invalid range"
        assert ranje[0] < ranje[1], "invalid range [1]"

        q = random.random()
        c = ceil(ranje[0] + q * (ranje[1] - ranje[0]))
        return int(c)

    def calculate(self, v):
        assert type(v) is int, "invalid value v"
        return int((self.multiplier * v + self.increment) % self.modulo)

    def __next__(self):
        q = self.vInt
        self.vInt = self.calculate(q)
        return q

#############------------------------------------------------------------

class CycleMap:

    '''
    '''
    def __init__(self, cRange:'int'):
        assert type(cRange) is int, "invalid type for cycle range"
        self.cRange = cRange
        self.mahp = OrderedDict()
        self.head = None
        self.phead = None
        return

    def set_map(self, m):
        assert type(m) is OrderedDict, "invalid type for map"
        assert CycleMap.is_valid_map(m), "invalid cycle map"
        assert len(m) == self.cRange, "invalid length for map"
        self.mahp = m

    @staticmethod
    def random_cycle_map(vortexRange):
        ks = np.arange(vortexRange)
        np.random.shuffle(ks)

        # select the head
        h = ks[0]
        h_ = h
        l = OrderedDict()
        rem = np.arange(vortexRange)
        for i in range(vortexRange - 1):
            possible = [r for r in rem if r != h]

            # choose random
            lp = len(possible)
            ri = random.randrange(lp)
            l[h] = possible[ri]

            rem = np.array(possible)
            h = l[h]

        l[h] = h_

        return l

    @staticmethod
    def is_valid_map(m):
        # fetch the first element in the map
        q = None
        for k in m.keys():
            q = k
            break

        # check for cycle
        x = [q]
        l = len(m)
        c = 0
        while c < l:
            r = m[x[-1]]
            x.append(r)
            c += 1

        # check that last element is first key
        if not (x[-1] == x[0]): return False
        # check that number of unique elements is l
        if len(set(x)) != l: return False
        return True

    def head_(self):
        for k,v in self.mahp.items():
            self.head = k
            self.c = k
            break

    def __next__(self):
        if type(self.c) == type(None): self.head_()
        q = self.mahp[self.c]
        self.c = q
        return q

    def v(self,k):
        return self.mahp[k]

    #### TODO: make generators for non-cycles

##########################################################################################

DELTA_ADD = lambda x,c: x + c
DELTA_MULT = lambda x,c: x * c
# TODO: future
DELTA_POLY = lambda x, poly: 1

'''
- rangesGanges: 2-tuple
- divider: 0.0|(float >= 1.0)

constant: 0 or 1
delta: divider n^-1
'''
def FloatDeltaGenerator_pattern1(rangesGanges, divider):
    assert is_valid_point(rangesGanges), "ranges ganges ranges ganges"
    assert divider >= 0.0, "invalid divider"

    # no delta
    if divider != 0.0:
        dividingLengthon = abs(rangesGanges[1] - rangesGanges[0]) / divider
        directororios = 1 if rangesGanges[1] > rangesGanges[0] else -1
        lengthonLengthon = dividingLengthon * directororios
        divider = lengthonLengthon
    return FloatDeltaGenerator(rangesGanges[0], DELTA_ADD, divider, rangesGanges[1])

#------------------------------------------------------------------
#### binary sequence generators

def generate_possible_binary_sequences(vecOrder, thisNumber, elements = [0,1]):

    if len(thisNumber) == vecOrder:
        yield thisNumber
        return

    q1, q2 = np.copy(thisNumber), np.copy(thisNumber)
    q1, q2 = np.hstack((q1,[elements[0]])), np.hstack((q1,[elements[1]]))
    yield from generate_possible_binary_sequences(vecOrder, q1,elements)
    yield from generate_possible_binary_sequences(vecOrder, q2,elements)

def generate_random_binary_sequence(vecOrder):
    assert type(vecOrder) is int, "invalid vec. order"
    return np.random.randint(0, 2, (vecOrder,))

####---------------------------------------------------------------------
#### uniform dist. numerical generators

"""
"""
def generate_uniform_sequence_in_bounds(vecOrder, bounds,rnd_struct=rng):
    assert is_2dmatrix(bounds), "invalid bounds {}".format(bounds)
    assert vecOrder == len(bounds) or len(bounds) == 1, "invalid bounds"

    if len(bounds) == 1:
        return rnd_struct.uniform(bounds[0,0], bounds[0,1], (vecOrder,))
    else:
        q = np.zeros((vecOrder,))
        for i in range(vecOrder):
            q[i] = rnd_struct.uniform(bounds[i,0], bounds[i,1])#, (vecOrder,))
        return q

def random_bounds_edge(bounds):
    assert is_bounds_vector(bounds), "invalid bounds"
    bs = generate_random_binary_sequence(bounds.shape[0])
    r = [i for i in range(bounds.shape[0])]
    return bounds[r,bs]

################################# TODO: noise methods need to be checked.


def k_random_points_in_bounds(minVec,maxVec,k):
    '''
    A method that generates `k` points in bounds [`minVec`,`maxVec`] by random.random().

    :param minVec: n-dimensional vector with each dimension the minumum value
    :type minVec: np.ndarray
    :param maxVec: n-dimensional vector with each dimension the maximal value
    :type maxVec: np.ndarray
    :param k: number of points to generate
    :type k: int, > 0
    '''
    assert np.all(maxVec - minVec >= 0.0), "invalid arguments min.,max. vec"

    d = maxVec - minVec
    for i in range(k):
        x = random.random()
        yield minVec + (d * x)

###------------------------

def one_random_noise(bounds,noiseRange):
    """
    :param bounds: bounds matrix
    :type bounds: np.ndarray, n x 2 matrix
    :param noiseRange: 2dmatrix, len is 1 or bounds.shape[0]
    """
    assert is_proper_bounds_vector(bounds), "invalid bounds"

    # set max distance for each dim.
    q = bounds[:,1] - bounds[:,0]
    us = generate_uniform_sequence_in_bounds(bounds.shape[0], noiseRange)
    q = q * us
    return q

def random_npoint_from_point(p1,l,roundDepth=5):
    p2 = np.copy(p1)
    r = np.arange(p1.shape[0])
    np.random.shuffle(r)
    r,x = r[:-1],r[-1]
    d = 0.0

    def delta_extreme():
        return math.sqrt(l ** 2 - d)

    def random_delta(j,randomize = True):
        delta = delta_extreme()
        delta = delta if random.random() > 0.5 else -delta

        if randomize:
            delta = random.uniform(0.0,delta)
        return delta

    for r_ in r:
        rd = random_delta(r_)
        p2[r_] = p2[r_] + rd
        d += (rd **2)
    p2[x] = p2[x] + random_delta(x,False)
    return np.round(p2,5)

def random_npoint_from_point_in_bounds_(b,p1,l,roundDepth = 5):
    """
    calculates a random n-dimensional point p2 in bounds `b` of distance `l`
    from `p1`.
    """

    p2 = np.copy(p1)
    r = np.arange(p1.shape[0])
    np.random.shuffle(r)
    r,x = r[:-1],r[-1]
    d = 0.0

    def delta_extreme():
        return math.sqrt(l ** 2 - d)

    def possible_extreme(j):
        return [b[j,0] - p1[j],b[j,1] - p1[j]]

    def random_delta(j,randomize = True):
        delta = delta_extreme()
        delta = delta if random.random() > 0.5 else -delta

        if randomize:
            delta = random.uniform(0.0,delta)

        pe = possible_extreme(j)
        pe = pe[1] if delta > 0 else pe[0]
        sols = [pe,delta]
        k = np.argmin(np.abs(sols))
        return sols[k]

    for r_ in r:
        rd = random_delta(r_)
        p2[r_] = p2[r_] + rd
        d += (rd **2)
    p2[x] = p2[x] + random_delta(x,False)
    return np.round(p2,5)

def random_npoint_from_point_in_bounds(b,p1,l,attempts = 15):
    t = False

    while not t and attempts > 0:
        p2 = random_npoint_from_point_in_bounds_(b,p1,l)
        #print("DISTANCE ", euclidean_point_distance(p1,p2))
        t = abs(euclidean_point_distance(p1,p2) - l) < 10 ** -5
        attempts -= 1
    return p2 if t else None

def add_noise_to_points_restricted_bounds(minVec,maxVec, points, noiseRange = np.array([[0.25,0.65]]), boundsRestriction = False):
    '''
    adds noise to points in restricted bounds by noise range,
    set to (0.25,0.65) by default.

    each output in yield is a point that may not lie within the bounds set by minVec
    and maxVec
    '''

    b = np.array([minVec,maxVec]).T
    for p in points:
        yield p + one_random_noise(b,noiseRange)

def random_noise_sequence(s,b,noiseRange):
    rn = []
    for i in range(s):
        n = one_random_noise(b,noiseRange)
        rn.append(n)
    return np.array(rn)

# TODO:
def generate_gaussian_sequence_in_bounds(mean, var):
    rng.normal()
    return -1
